import os
from unittest.mock import MagicMock

import pytest
from lmwrapper.openai_wrapper import OpenAiModelNames, get_open_ai_lm
from lmwrapper.structs import LmPrediction

from synthegrator.code_problems import (
    CodeProblem,
    CodeSolution,
    DiscoveredTestsuite,
)
from synthegrator.code_solver import LmCodeSolverAutoRegressive
from synthegrator.environments import CodeExecutionEnvironment, ProjectWorkingDirectory
from synthegrator.lang_specs.lang_spec_java import JavaLangSpec
from synthegrator.memory_fs import ProjectDir
from synthegrator.problem_rendering import (
    render_single_edit_codeproblem_autoregressive,
)
from synthegrator.sandboxing import JAVA_DEFAULT_DOCKER_ENV
from synthegrator.solution_eval import evaluate_code_problem_execution
from synthegrator.transformation_spec import (
    SolveStep,
    StsInsert,
    StsSelectorCharSpan,
    StsSpecStatement,
    TransformationSpec,
)

IN_GITHUB_ACTIONS = os.getenv("GITHUB_ACTIONS") == "true"


def get_java_double_problem():
    content = (
        # fmt: off
        "import java.util.*;\n"
        "import java.lang.*;\n"
        "\n"
        "class Solution {\n"
        "    /**\n"
        "    Double the input number.\n"
        "    */\n"
        "    public int doubleNumber(int x) {\n"
        "}\n"
        # fmt: on
    )
    content_test = (
        # fmt: off
        "import org.junit.jupiter.api.Test;\n"
        "import static org.junit.jupiter.api.Assertions.assertEquals;\n"
        "\n"
        "class SolutionTest {\n"
        "\n"
        "    @Test\n"
        "    void testDouble() {\n"
        "        Solution solution = new Solution();\n"
        "\n"
        "        assertEquals(4, solution.doubleNumber(2));\n"
        "        assertEquals(0, solution.doubleNumber(0));\n"
        "        assertEquals(-6, solution.doubleNumber(-3));\n"
        "        assertEquals(20, solution.doubleNumber(10));\n"
        "    }\n"
        "}\n"
        # fmt: on
    )
    select_char = len(content) - len("}\n")
    problem = CodeProblem(
        instructions="Complete this method",
        working_directory=ProjectWorkingDirectory(
            files=ProjectDir.construct_with_files(
                [
                    ("src/main/java/Solution.java", content),
                    ("src/test/java/SolutionTest.java", content_test),
                ],
            ),
        ),
        transformation_spec=TransformationSpec(
            statements=(
                StsSpecStatement(
                    "src/main/java/Solution.java",
                    StsSelectorCharSpan(select_char, select_char),
                    StsInsert(
                        lang_spec_name="java",
                        set_base_space_indent=8,
                        stop_at_block_end=True,
                        set_ending_text="}",
                    ),
                ),
            ),
        ),
        environment=CodeExecutionEnvironment(JAVA_DEFAULT_DOCKER_ENV),
        instructions_are_essential=False,
        test_cases=[
            DiscoveredTestsuite(
                cmds=[
                    "mvn test",
                    "cat target/surefire-reports/TEST-SolutionTest.xml",
                ],
            ),
        ],
    )
    render = render_single_edit_codeproblem_autoregressive(problem)
    assert render.prompt.text == (
        # fmt: off
        "import java.util.*;\n"
        "import java.lang.*;\n"
        "\n"
        "class Solution {\n"
        "    /**\n"
        "    Double the input number.\n"
        "    */\n"
        "    public int doubleNumber(int x) {\n"
    )
    problem.known_solutions.append(
        CodeSolution(
            solve_steps=(
                SolveStep(
                    path=render.path,
                    mark_id=render.changed_element.mark_id,
                    value=" " * 8 + "return x * 2;\n    }",
                ),
            ),
            problem=problem,
            lang_spec=JavaLangSpec(),
        ),
    )
    return problem


def test_simple_method_problem_known_solution():
    problem = get_java_double_problem()
    eval = evaluate_code_problem_execution(problem, problem.known_solutions[0])
    assert eval.all_passed()


@pytest.mark.skipif(IN_GITHUB_ACTIONS, reason="No API key")
@pytest.mark.skip(reason="Need to sort to through include lm things")
def test_simple_method_problem_gpt():
    problem = get_java_double_problem()
    lm = get_open_ai_lm(
        OpenAiModelNames.gpt_3_5_turbo_instruct,
    )
    solver = LmCodeSolverAutoRegressive(lm, cache_lm=False, include_lm_response=True)
    solution = solver.solve(problem)
    print("prompt", solution.lm_prediction.prompt)
    print("completion text", solution.lm_prediction.completion_text)
    print(solution.solve_steps[0])
    eval = evaluate_code_problem_execution(problem, solution)
    assert eval.all_passed()
    # Disable lm resp
    solver = LmCodeSolverAutoRegressive(lm, cache_lm=False, include_lm_response=False)
    solution = solver.solve(problem)
    assert solution.lm_prediction is None


@pytest.mark.skipif(IN_GITHUB_ACTIONS, reason="No API key")
def test_simple_method_problem_gpt_chat():
    problem = get_java_double_problem()
    lm = get_open_ai_lm(OpenAiModelNames.gpt_3_5_turbo)
    solver = LmCodeSolverAutoRegressive(lm)
    solution = solver.solve(problem)
    print(solution.solve_steps[0])
    eval = evaluate_code_problem_execution(problem, solution)
    assert eval.all_passed()


@pytest.mark.parametrize(
    "gen_return",
    [
        "    return x * 2;\n    }\n",
        "    return x * 2;\n    }\n} class ExtraCode() {}\n int x =",
        "    return x * 2;\n",
    ],
)
def test_simple_method_problem_mocked(gen_return: str):
    problem = get_java_double_problem()
    lm = get_open_ai_lm(OpenAiModelNames.gpt_3_5_turbo)
    lm.predict = MagicMock(
        return_value=LmPrediction(
            gen_return,
            None,
            None,
        ),
    )
    solver = LmCodeSolverAutoRegressive(lm)
    solution = solver.solve(problem)
    print(solution.solve_steps[0])
    assert "ExtraCode" not in solution.solve_steps[0].value
    eval = evaluate_code_problem_execution(problem, solution)
    assert eval.all_passed()


@pytest.mark.parametrize(
    "gen_return",
    [
        "    return x * 2;\n    }\n",
        "    return x * 2;\n    }",
        "    return x * 2;\n    }\n} class ExtraCode() {}",
        "    return x * 2;\n    }\n void ExtraCode() {}\n}",
    ],
)
def test_truncate_func(gen_return):
    lang_spec = JavaLangSpec()
    out = lang_spec.truncate_function_completion(gen_return)
    assert out == "    return x * 2;\n    }"


@pytest.mark.parametrize(
    "gen_return",
    [
        "    return x * 2;}",
        "    return x * 2;}\n}",
    ],
)
def test_truncate_func_2(gen_return):
    lang_spec = JavaLangSpec()
    out = lang_spec.truncate_function_completion(gen_return)
    assert out == "    return x * 2;}"


@pytest.mark.parametrize(
    "gen_return",
    [
        "        int x = 3;\n        return x * 2;\n    } class ExtraCode() {}",
    ],
)
def test_truncate_func_3(gen_return):
    lang_spec = JavaLangSpec()
    out = lang_spec.truncate_function_completion(gen_return)
    assert out == "        int x = 3;\n        return x * 2;\n    }"


def test_truncate_func_loop():
    lang_spec = JavaLangSpec()
    completion = (
        # fmt: off
        "        int x = 3;\n"
        "        for (int i = 0; i < 10; i++) {\n"
        "            x += i;\n"
        "        }\n"
        "        return x * 2;\n"
        "    }\n"
        "}\n"
        "class ExtraCode() {}"
        # fmt: on
    )
    out = lang_spec.truncate_function_completion(completion)
    assert out == "\n".join(completion.splitlines()[:-2])


def test_truncate_func_inner_func():
    lang_spec = JavaLangSpec()
    completion = (
        # fmt: off
        "        int x = 3;\n"
        "        void bar(int x) {\n"
        "            x += 1;\n"
        "        }\n"
        "        return x * 2;\n"
        "    }\n"
        "}\n"
        "class ExtraCode() {}"
        # fmt: on
    )
    out = lang_spec.truncate_function_completion(completion)
    assert out == "\n".join(completion.splitlines()[:-2])


# def test_simple_method_problem_codet5():
#    problem = _get_problem()
#    lm = get_huggingface_lm("Salesforce/codet5p-220m")
#    solver = LmCodeSolverAutoRegressive(lm)
#    solution = solver.solve(problem, lang_spec=JavaLangSpec())
#    print(solution.solve_steps[0])
#    eval = evaluate_code_problem(problem, solution)
#    assert eval.all_passed()



"""
Dataclasses and helper functions for defining a
problem want to synthesize code for.
"""

import inspect
from abc import ABC, abstractmethod
from collections.abc import Callable, Sequence
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, NamedTuple

import xxhash
from lmwrapper.structs import LmPrediction, LmPrompt

from synthegrator.environments import (
    CodeExecutionEnvironment,
    ProjectWorkingDirectory,
)
from synthegrator.lang_specs.lang_spec import (
    LangSpec,
)
from synthegrator.lang_specs.lang_spec_java import JavaLangSpec
from synthegrator.lang_specs.lang_spec_python import (
    PythonLangSpec,
    assert_spec_is_python,
)
from synthegrator.memory_fs import ProjectDir
from synthegrator.sandboxing import (
    Cmd,
    CmdExecResult,
    TestsuiteExecResult,
    convert_pytest_outputs_to_testsuite_exec_result,
)
from synthegrator.transformation_spec import (
    MarkElement,
    SolveStep,
    StsValueQuery,
    TransformationSpec,
)
from synthegrator.uncertainty_modeling import DelayedUncertaintyEstimate

if TYPE_CHECKING:
    # A seeming hack to be able to import the SolutionEvaluator during
    #   the typing of CodeProblem without a circular import. Not quite
    #   how good of a solution this will be.
    from synthegrator.solution_eval import SolutionEvaluator


class TestCaseFormattedValue(NamedTuple):
    call_expression: str
    setup: str


@dataclass
class TestCase(ABC):
    __test__ = False
    is_hidden_test: bool
    """In a coding problem some test case might be hidden from the solver"""
    fail_message: str

    @abstractmethod
    def get_hash(self) -> str:
        hash_calc = xxhash.xxh64()
        if self.fail_message:
            hash_calc.update(self.fail_message)
        hash_calc.update(str(self.is_hidden_test))
        return hash_calc.hexdigest()

    def __post_init__(self):
        pass

    @abstractmethod
    def result_is_success(self, output: CmdExecResult) -> bool:
        # Deprecated with docker running
        raise NotImplementedError

    @abstractmethod
    def format_for_test_framework(
        self,
        lang_spec: LangSpec,
        override_test_id: str | None = None,
    ) -> str:
        """Formats for running on a test framework like pytest."""
        raise NotImplementedError


def _pytest_with_body(test_body: str, test_id, fixtures=None) -> str:
    if not fixtures:
        fixtures = []
    # Make a random id for the test
    test_body = "\n".join("    " + line for line in test_body.split("\n"))
    return f"def test_{test_id}({', '.join(fixtures)}):\n{test_body}"


@dataclass
class TestCaseExpressionIsEq(TestCase):
    __test__ = False
    """
    Checks whether calling expression equals a value. Currently implemented
    only for python
    """

    input_expression: str
    output: Any
    lang_spec: LangSpec

    def get_hash(self) -> str:
        hash_calc = xxhash.xxh64()
        hash_calc.update(super().get_hash())
        hash_calc.update(self.input_expression)
        hash_calc.update(str(self.output))
        hash_calc.update(self.lang_spec.get_lang_md_name())
        return hash_calc.hexdigest()

    def __post_init__(self):
        assert_spec_is_python(self.lang_spec)

    def result_is_success(self, output: CmdExecResult) -> bool:
        return output.output == self.output

    def format_for_test_framework(
        self,
        lang_spec: LangSpec,
        override_test_id: str | None = None,
    ) -> str:
        assert_spec_is_python(lang_spec)
        return _pytest_with_body(
            f"assert {self.input_expression} == {self.output}",
            override_test_id or self.get_hash(),
        )


@dataclass
class TestCaseMethodCallIsEq(TestCase):
    __test__ = False
    method_name: str
    test_input: tuple[Any, ...]
    output: Any

    def get_hash(self) -> str:
        hash_calc = xxhash.xxh64()
        hash_calc.update(super().get_hash())
        hash_calc.update(str(self.test_input))
        hash_calc.update(str(self.output))
        return hash_calc.hexdigest()

    def _format_method_call(
        self,
        lang_spec: LangSpec,
    ) -> str:
        assert_spec_is_python(lang_spec)
        return f"{self.method_name}({', '.join(map(str, self.test_input))})"

    def result_is_success(self, output: CmdExecResult) -> bool:
        return output.output == self.output

    def format_for_test_framework(
        self,
        lang_spec: LangSpec,
        override_test_id=None,
    ):
        assert_spec_is_python(lang_spec)
        return _pytest_with_body(
            f"assert {self._format_method_call(lang_spec)} ==" f" {self.output}",
            override_test_id or self.get_hash(),
        )


@dataclass
class TestCaseArbitraryMethod(TestCase):
    method_body: str
    lang_spec: LangSpec
    fixtures: tuple[str, ...] = ()

    def get_hash(self) -> str:
        hash_calc = xxhash.xxh64()
        hash_calc.update(super().get_hash())
        hash_calc.update(str(self.method_body))
        hash_calc.update(
            str(self.lang_spec.get_lang_md_name()),
        )  # TODO: add other info from spec
        hash_calc.update(str(self.fixtures))
        return hash_calc.hexdigest()

    def result_is_success(self, output: CmdExecResult) -> bool:
        return output.completed

    def format_for_test_framework(
        self,
        lang_spec: LangSpec,
        override_test_id: str | None = None,
    ) -> str:
        assert_spec_is_python(lang_spec)
        return _pytest_with_body(
            self.method_body,
            override_test_id or self.get_hash(),
            self.fixtures,
        )


class DiscoveredTestsuite(TestCase):
    """
    A test case for when running a suite that already exists in the environment.
    It specifies the necessary commands to get the valid suite xml.
    """

    def __init__(
        self,
        cmds: Sequence[Cmd | str],
        cmds_results_to_suite: (
            Callable[[Sequence[CmdExecResult]], TestsuiteExecResult] | None
        ) = None,
    ):
        self.is_hidden_test = True
        self.fail_message = f"Fail discovered test {cmds}"
        self.cmds = Cmd.from_cmds_str_mix(cmds)
        self.cmds_results_to_suite = cmds_results_to_suite
        if self.cmds_results_to_suite is None:
            if len(self.cmds) < 2:
                msg = "Cmds length invalid. Failed run?"
                raise Exception(msg)

            if "cat " not in self.cmds[-1].cmd:
                msg = "Cat command not executed. Failed run?"
                raise Exception(msg)

            self.cmds_results_to_suite = self._default_cmds_results_suite

    def get_hash(self) -> str:
        hash_calc = xxhash.xxh64()
        hash_calc.update(super().get_hash())
        hash_calc.update(str(self.cmds))
        hash_calc.update(inspect.getsource(self.cmds_results_to_suite))
        return hash_calc.hexdigest()

    @staticmethod
    def _default_cmds_results_suite(
        cmds_results: Sequence[CmdExecResult],
    ) -> TestsuiteExecResult:
        return convert_pytest_outputs_to_testsuite_exec_result(
            cmds_results[-2],
            cmds_results[-1],
        )

    def result_is_success(self, output: CmdExecResult) -> bool:
        return output.completed

    def format_for_test_framework(
        self,
        lang_spec: LangSpec,
        override_test_id: str | None = None,
    ) -> str:
        raise NotImplementedError


@dataclass
class CodeSolution:
    solve_steps: Sequence[SolveStep]
    problem: "CodeProblem"
    lang_spec: LangSpec = None  # TODO remove this eventually
    time: float = None
    """Time in seconds of how long the solver took to produce the solution."""

    def __post_init__(self):
        if not isinstance(self.solve_steps, Sequence):
            msg = "Solve steps must be a Sequence"
            raise TypeError(msg)
        if not all(isinstance(step, SolveStep) for step in self.solve_steps):
            msg = "All solve steps must be a SolveStep"
            raise TypeError(msg)

    def apply(self) -> ProjectDir:
        from synthegrator.solution_eval import apply_solution

        return apply_solution(self)

    def pull_value_answers(self) -> list[tuple[MarkElement[StsValueQuery], str]]:
        from synthegrator.solution_eval import pull_value_answers_from_solution

        return pull_value_answers_from_solution(self)

    def dict_serialize(self) -> dict:
        out = {
            "solve_steps": sorted(
                [step.dict_serialize() for step in self.solve_steps],
                key=lambda x: (x["path"], x["mark_id"]),
            ),
            "problem": self.problem.dict_serialize(),
            "time": self.time,
            "lang_spec": self.lang_spec.dict_serialize() if self.lang_spec else None,
            "number_of_solve_steps": len(self.solve_steps),
        }
        if len(self.solve_steps) == 1:
            out["only_solve_value"] = self.solve_steps[0].value
            out["only_solve_path"] = self.solve_steps[0].path
        return out


@dataclass
class LmCodeSolution(CodeSolution):
    uncertainty_estimates: list[DelayedUncertaintyEstimate] = None
    prompt: LmPrompt | None = None
    lm_prediction: LmPrediction | None = None

    def __post_init__(self):
        super().__post_init__()
        if self.uncertainty_estimates is None:
            self.uncertainty_estimates = []

    def dict_serialize(
        self,
        remove_redundant_prompt: bool = True,
    ) -> dict:
        out = super().dict_serialize()
        # out["uncertainty_estimates"] = [
        #    est.dict_serialize()
        #    for est in self.uncertainty_estimates
        # ]
        if self.prompt:
            out["prompt"] = self.prompt.dict_serialize()
        if self.lm_prediction:
            out["lm_prediction"] = self.lm_prediction.dict_serialize()
            if remove_redundant_prompt and self.prompt:
                # There will also be the prompt in the response which
                # is redundant
                del out["lm_prediction"]["prompt"]
        return out


DEFAULT_CODE_ENV = CodeExecutionEnvironment()
DEFAULT_LANG_SPECS = (
    (".py", PythonLangSpec()),
    (".java", JavaLangSpec()),
)


@dataclass
class CodeProblem:
    instructions: str = None
    """The main instructions describing what we want a solver to do in this problem"""
    working_directory: ProjectWorkingDirectory = None
    """The working directory that contains the relevant files for the problem.
    Optionally, the files can exist already in the execution environment, but then
    they are opaque to the solvers."""
    transformation_spec: TransformationSpec = None
    """Outlines what a solver is and isn't allowed or expected to edit"""
    test_cases: list[TestCase] = field(default_factory=list)
    known_solutions: list[CodeSolution] = field(default_factory=list)
    environment: CodeExecutionEnvironment = DEFAULT_CODE_ENV
    dataset_name: str = None
    """Labeling which dataset this problem came from"""
    problem_id: str = None
    """Some sort of identifier within the dataset. The pair of (dataset_name, problem_id)
    should be unique."""
    instructions_are_essential: bool = True
    """Whether the instructions include key information for solving the problem.
    For some method or line completion tasks or when the task is embedded into
    documents in the code, the instructions might not be essential. However
    if you are generating code with no documentation or other guide on the task,
    then the instructions might be essential. This can influence the behaviour
    of autoregressive models which might perform better when the instructions
    are naturally integrated into the code."""
    extension_to_lang_specs: tuple[tuple[str, LangSpec], ...] = DEFAULT_LANG_SPECS
    """A tuple that contains all the language specs that are used in the problem.
    Each element is a tuple of (extension, lang_spec). Paths that end with the
    extension will use the associate language spec"""
    preferred_solution_evaluator: "SolutionEvaluator" = None
    """Preferred solution evaluator for the problem. Can be used in case certain
     problems use metrics. If set to `None` then a SolutionEvaluator.
    will be used if there are test cases. Otherwise it will be None."""
    instructable_metadata: dict[str, Any] = None
    """NOT IMPLEMENTED. Intended for things that certain solvers might using
    while solving, but might do it in their own specially formatted way. Some
    examples might include expected commit message or something. Several of
    the standard metadata (eg, a commit message, issue text and discussion, etc)
    are intended to be standardized with certain keys (TODO)."""
    noninstructable_metadata: dict[str, Any] = None
    """Intended for extra problem metadata (for example which split a dataset example
    came from or other info) which is not intended for use by the solver but
    just for processing after the fact."""
    state_info = None
    """NOT IMPLEMENTED. intended to package extra external information available when
    solving the problem (like traces, error messages, etc.). This still needs
    to be further designed out and is not implemented."""
    past_solve_context: list[CodeSolution] = None
    """Context from some iterations trying to solve this"""

    def __post_init__(self):
        if self.state_info is not None:
            raise NotImplementedError
        if not isinstance(self.known_solutions, list):
            msg = "known_solutions must be a list"
            raise TypeError(msg)
        if not all(
            isinstance(s, CodeSolution) for s in self.known_solutions
        ) or isinstance(self.known_solutions, CodeSolution):
            msg = "known_solutions must be a list of CodeSolution"
            raise TypeError(msg)
        if self.preferred_solution_evaluator is None and (
            self.test_cases
            or len(list(self.transformation_spec.classification_question_statements()))
            > 0
        ):
            from synthegrator.solution_eval import SolutionEvaluator

            self.preferred_solution_evaluator = SolutionEvaluator.default_instance()

    def get_lang_spec_for_path(self, path: str) -> LangSpec | None:
        for ext, lang_spec in self.extension_to_lang_specs:
            if path.endswith(ext):
                return lang_spec
        return None

    def get_non_hidden_test_cases(self) -> list[TestCase]:
        return [t for t in self.test_cases if not t.is_hidden_test]

    def dict_serialize(self, just_serialize_identifiers: bool = True) -> dict:
        out = {
            "dataset_name": self.dataset_name,
            "problem_id": self.problem_id,
        }
        if not just_serialize_identifiers:
            raise NotImplementedError
        return out

    def __eq__(self, other):
        if not isinstance(other, CodeProblem):
            return False
        # We basically compare everything except:
        #   - known_solutions since they might reference back to this problem
        return (
            self.problem_id == other.problem_id
            and self.instructions == other.instructions
            and self.working_directory == other.working_directory
            and self.transformation_spec == other.transformation_spec
            and self.test_cases == other.test_cases
            and self.environment == other.environment
            and self.dataset_name == other.dataset_name
            and self.instructions_are_essential == other.instructions_are_essential
            and self.extension_to_lang_specs == other.extension_to_lang_specs
            and self.past_solve_context == other.past_solve_context
        )



from collections import Counter
import dataclasses
from itertools import chain, islice
from pprint import pprint

from synthegrator.code_problems import CodeProblem, TestCaseArbitraryMethod
from synthegrator.code_problem_builders import make_simple_file_generation, \
    make_simple_method_completion_problem
from synthegrator.lang_specs.lang_spec_python import PythonLangSpec
from synthegrator.sandboxing import PY_DEFAULT_DOCKER_ENV, ExecLimits
from synthegrator.synthdatasets.apps import apps_assertion_code_gen
from synthegrator.synthdatasets.livecodebench.authorcode.benchmarks.code_generation import \
    load_code_generation_dataset, TestType, Difficulty, LivecodeInternalCodeGenerationProblem
from synthegrator.util import IteratorWithLength

LIVECODEBENCH_CODE_ENV = (
    "import sys\nimport time\nimport itertools\nfrom itertools import accumulate,"
    " product, permutations, combinations\nimport collections\nfrom collections import"
    " Counter, OrderedDict, deque, defaultdict, ChainMap\nfrom functools import"
    " lru_cache\nimport math\nfrom math import sqrt, sin, cos, tan, ceil, fabs, floor,"
    " gcd, exp, log, log2\nimport fractions\nfrom typing import List, Tuple\nimport"
    " numpy as np\nimport random\nimport heapq\nfrom heapq import *\n"
)


def _create_stdin_test_case(input_data: str, expected_output: str, index: int, is_hidden: bool) -> TestCaseArbitraryMethod:
    """
    Create a test case for STDIN/STDOUT testing.
    
    Args:
        input_data: The input data to provide via stdin
        expected_output: The expected output to stdout
        index: Test case index (for naming)
        is_hidden: Whether this is a hidden test case
        
    Returns:
        A TestCaseArbitraryMethod instance configured for stdin/stdout testing
    """
    def outter_shorten_repr(d):
        if len(repr(d)) > 100:
            return repr(d)[:150] + '...'
        return repr(d)
    short_input_data = outter_shorten_repr(input_data)
    short_expected_output = outter_shorten_repr(expected_output)

    return TestCaseArbitraryMethod(
        is_hidden_test=is_hidden,
        fail_message=f"Test case {index+1} failed",
        method_body=f"""
# LCB STDIN TEST CASE
import sys
import io
import os
from unittest.mock import patch

# Capture stdin/stdout
stdin_data = {repr(input_data)}
expected_output = {repr(expected_output)}

# Find the solution file
solution_path = os.path.join(os.getcwd(), 'solution.py')
assert os.path.exists(solution_path), "solution.py file does not exist"

# Create a patched stdin and capture stdout
#input_buffer = io.StringIO(stdin_data)
input_buffer = io.StringIO(stdin_data if stdin_data.endswith('\\n') else stdin_data+'\\n')
output_buffer = io.StringIO()

with patch('sys.stdin', input_buffer), patch('sys.stdout', output_buffer):
    # Execute the code directly instead of importing
    import solution
    solution.main()
    #with open(solution_path, 'r') as f:
    #    solution_code = f.read()
    #    exec(solution_code)
    
    # Get the captured output
    actual_output = output_buffer.getvalue()
    
    # Handle the case of single-line output without newlines
    if not actual_output.strip():
        actual_lines = []
    elif '\\n' not in actual_output:
        actual_lines = [actual_output.strip()]
    else:
        actual_lines = [line.rstrip() for line in actual_output.splitlines()]
    
    if not expected_output.strip():
        expected_lines = []
    elif '\\n' not in expected_output:
        expected_lines = [expected_output.strip()]
    else:
        expected_lines = [line.rstrip() for line in expected_output.splitlines()]
    
    # Simple string format to avoid variable reference issues
    if actual_lines != expected_lines:
        msg = f'Expected output does not match actual output.\\n'
        actual_lines_repr = repr(actual_lines)
        if len(actual_lines_repr) > 150:
            actual_lines_repr = actual_lines_repr[:150] + '...'
        def shorten(d):
            if len(repr(d)) > 150:
                return repr(d)[:150-1] + '...'
            return repr(d)
        msg += f'Input: {{shorten(stdin_data)}},\\nGot: {{shorten(actual_lines)}},\\nExpected: {{shorten(expected_lines)}}'
        raise AssertionError(msg)
""",
        lang_spec=PythonLangSpec(),
        fixtures=("monkeypatch", "capfd"),
    )


LIVECODEBENCH_DOCKER_ENV = dataclasses.replace(
    PY_DEFAULT_DOCKER_ENV,
    default_limits=ExecLimits(
        # The default seems to be about 6 seconds
        # https://github.com/LiveCodeBench/LiveCodeBench/blob/effd1e15b37cf0135e6aaceabc9ae0dc61a5e838/lcb_runner/runner/parser.py#L93
        # We'll give it a bit more time
        timeout_cpu_s=10,
        timeout_realtime_s=20,  # A bit more since might be multithreaded.
    )
)


def _produce_livecode_ex(ex: LivecodeInternalCodeGenerationProblem) -> CodeProblem:
    """
    Produce a CodeProblem for LiveCodeBench examples with STDIN test cases.
    
    Args:
        ex: A CodeGenerationProblem instance from the livecodebench dataset
        
    Returns:
        A CodeProblem instance
    
    Raises:
        ValueError: If not all test cases are of type STDIN
    """
    all_test_cases = list(chain(ex.private_test_cases, ex.public_test_cases))

    test_cases = []
    if all(tc.testtype == TestType.FUNCTIONAL for tc in all_test_cases):
        solution_start_good = ex.starter_code.strip().startswith("class Solution:")
        solution_lines_good = ex.starter_code.strip().count("\n") < 4
        if not (solution_start_good and solution_lines_good):
            raise ValueError("Starter code must be a simple function/class definition:\n" + ex.starter_code)
        signature = ex.starter_code.strip().split("\n")[1].strip()
        signature = signature.replace("self, ", "")
        func_name = ex.metadata["func_name"]
        for i, tc in enumerate(all_test_cases):
            is_hidden = tc in ex.private_test_cases
            test_case = _create_func_call_tc(tc.input, tc.output, i, is_hidden, func_name)
            test_cases.append(test_case)
        override_instructions = None
    else:
        if not all(tc.testtype == TestType.STDIN for tc in all_test_cases):
            raise ValueError(f"All test cases must be of type STDIN, got {set(tc.testtype for tc in all_test_cases)}")
        signature = "def main() -> None:"
        func_name = "main"
        for i, tc in enumerate(all_test_cases):
            is_hidden = tc in ex.private_test_cases
            test_case = _create_stdin_test_case(tc.input, tc.output, i, is_hidden)
            test_cases.append(test_case)
        override_instructions = (
            "Complete the following method. "
            "The solution requires reading from standard input and writing to standard output. "
            "You can do this with the input() and print() functions, respectively. "
            "Give only the solution body that can go inside `def main() -> None`. It is fine and recommended "
            "to have any needed imports inside the function. Do not include the `def main` function "
            "signature. Do not repeat the instructions docstring."
        )
    question_content = ex.question_content.replace("\r", "")
    return make_simple_method_completion_problem(
        prompt=f"{ex.question_title}\n\n{question_content}",
        signature=signature,
        test_cases=test_cases,
        override_instructions=override_instructions,
        #model_context=ex.starter_code,
        dataset_name="livecodebench",
        execution_context_hidden_from_model=LIVECODEBENCH_CODE_ENV,
        problem_id=f"{ex.platform.value}_{ex.question_id}",
        known_solutions=[],  # No known solutions for now
        docker_env=LIVECODEBENCH_DOCKER_ENV,
    )
    # Create the problem using the builder function
    return make_simple_file_generation(
        prompt=f"{ex.question_title}\n\n{ex.question_content}",
        test_cases=test_cases,
        model_context=ex.starter_code,
        dataset_name="livecodebench",
        problem_id=f"{ex.platform.value}_{ex.question_id}",
        known_solutions=[],  # No known solutions for now
    )


def _create_func_call_tc(
    input_data: str,
    expected_output: str,
    index: int,
    is_hidden: bool,
    func_name: str,
) -> TestCaseArbitraryMethod:
    """
    Create a test case for functional testing.

    Args:
        input_data: The input data to provide to the function
        expected_output: The expected output from the function
        index: Test case index (for naming)
        is_hidden: Whether this is a hidden test case
        func_name: The name of the function to call

    Returns:
        A TestCaseArbitraryMethod instance configured for functional testing
    """
    input_str = _lcb_assertion_code_gen(input_data)
    return TestCaseArbitraryMethod(
        is_hidden_test=is_hidden,
        fail_message=f"Test case {index+1} failed",
        method_body=f"assert {func_name}({input_str}) == {expected_output}",
        lang_spec=PythonLangSpec(),
    )


def _lcb_assertion_code_gen(x: str) -> str:
    if type(x).__name__ != "str":
        raise ValueError(f"Unsupported type: {type(x).__name__}")
    return ", ".join(x.split("\n"))
    match str(type(x).__name__):
        case "str":
            y = x
        case "list":
            y = str(x)
        case "bool":
            y = str(x)
        case "int":
            y = str(x)
        case _:
            msg = "Unsupported type"
            raise Exception(msg, type(x))
    return y


def yield_livecode_problems(
    filter_difficulty: Difficulty | str = None,
    filter_test_type: TestType = None,
    version_tag: str = "release_v6",
    max_problems: int = None,
):
    filter_difficulty = (
        Difficulty(filter_difficulty) if isinstance(filter_difficulty, str) else filter_difficulty
    )
    def itr():
        num_yield = 0
        for ex in load_code_generation_dataset(version_tag):
            if max_problems and num_yield >= max_problems:
                break
            if filter_difficulty and ex.difficulty != filter_difficulty:
                continue
            if filter_test_type:
                all_test_cases = list(chain(ex.private_test_cases, ex.public_test_cases))
                if not all(tc.testtype == filter_test_type for tc in all_test_cases):
                    continue
            prob = _produce_livecode_ex(ex)
            yield prob
            num_yield += 1
    return IteratorWithLength(itr(), max_problems)


def explore():
    """Print counts of different test case types in the livecodebench dataset."""
    ct = Counter()
    for ex in load_code_generation_dataset():
        for tc in chain(ex.private_test_cases, ex.public_test_cases):
            ct[tc.testtype] += 1
    print(ct)


if __name__ == "__main__":
    explore()



from synthegrator.code_problem_builders import make_simple_method_completion_problem
from synthegrator.few_shotting import Bm25FewShotLibrary


def test_bm25_few_shot_library():
    def make_prob(text):
        return make_simple_method_completion_problem(
            prompt=text, signature="def foo(x):", test_cases=[]
        )

    problems = [
        make_prob("Return the average of a list of numbers."),
        make_prob("Return the sum of a list of numbers."),
        make_prob("Return the maximum of a list of numbers."),
        make_prob("Return the minimum of a list of numbers."),
        make_prob("Return the median of a list of numbers."),  # -5
        make_prob("Return the median of a list."),  # -4
        make_prob("Return the median of a list of."),  # -3
        make_prob("Return a number that makes happy"),  # -2
        make_prob("Return a number that makes very happy"),  # -1
    ]
    library = Bm25FewShotLibrary(problems)
    sample = library.sample_problems(1, problems[-1])
    assert len(sample) == 1
    assert sample[0] == problems[-2]
    print(sample[0].working_directory.files.get_only_file().content_str)

    library = Bm25FewShotLibrary(problems)
    sample = library.sample_problems(2, problems[-4])
    assert len(sample) == 2
    assert sample[0] in (problems[-3], problems[-5])
    assert sample[1] in (problems[-3], problems[-5])



import shutil
from pathlib import Path

import pandas as pd

Path("Defects4J_projects_clean").mkdir(exist_ok=True)

df = pd.read_csv(
    "Defects4J_oneLiner_metadata.csv",
    header=None,
    names=["Project", "ID", "File_Path", "LOC"],
)


DEFECTS4J_COMMAND = "defects4j"

for _, row in df.iterrows():
    project_name = row["Project"]
    project_id = str(row["ID"])
    file_path = str(row["File_Path"])

    buggy_path = Path(
        f"/Users/claudio/Projects/synthegrator/Defects4J_projects/{project_name}_{project_id}/{file_path}",
    )
    fixed_path = Path(
        f"/Users/claudio/Projects/synthegrator/Defects4J_projects/{project_name}_{project_id}_fixed/{file_path}",
    )

    buggy_path_clean = Path(
        f"Defects4J_projects_clean/{project_name}_{project_id}/{file_path}",
    )
    fixed_path_clean = Path(
        f"Defects4J_projects_clean/{project_name}_{project_id}_fixed/{file_path}",
    )
    buggy_path_clean.parent.mkdir(exist_ok=True, parents=True)
    fixed_path_clean.parent.mkdir(exist_ok=True, parents=True)
    if buggy_path.exists() and fixed_path.exists():
        shutil.copy(buggy_path, buggy_path_clean)
        shutil.copy(fixed_path, fixed_path_clean)
    else:
        print("Warning! Files don't exist")
        print("Buggy", buggy_path.exists(), buggy_path)
        print("Fixed", fixed_path.exists(), fixed_path)
